# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import connectome_manipulator.connectome_comparison.properties as test_module


def test_null():
    pass
